"""Tests for the GraphQL interface for gbp-fl"""

from dataclasses import replace
from unittest import TestCase
from unittest.mock import Mock

import gbp_testkit.fixtures as testkit
from django.test import TestCase as DjangoTestCase
from gbp_testkit.helpers import graphql
from gentoo_build_publisher import publisher
from gentoo_build_publisher.graphql import schema
from gentoo_build_publisher.records import BuildRecord
from gentoo_build_publisher.types import Build as GBPBuild
from unittest_fixtures import Fixtures, given, where

from gbp_fl.types import BinPkg, Build

from . import lib

# pylint: disable=missing-docstring


@given(lib.repo, lib.bulk_content_files, testkit.client)
@where(records_backend="memory", repo="gbp_fl.graphql.queries.repo")
class FileListSearchTests(TestCase):
    def test_search_without_machine(self, fixtures: Fixtures) -> None:
        f = fixtures
        repo = f.repo
        repo.files.bulk_save(f.bulk_content_files)

        query = """
          query filesStartingWithBa {
            flSearch(key: "ba*") { path binpkg { cpvb } }
          }
        """
        result = graphql(fixtures.client, query)

        self.assertTrue("errors" not in result, result.get("errors"))
        self.assertEqual(len(result["data"]["flSearch"]), 4)

    def test_search_without_machine_no_match(self, fixtures: Fixtures) -> None:
        query = 'query { flSearch(key: "python") { path binpkg { cpvb } } }'

        result = graphql(fixtures.client, query)

        self.assertTrue("errors" not in result, result.get("errors"))
        self.assertEqual(len(result["data"]["flSearch"]), 0)

    def test_search_with_machine(self, fixtures: Fixtures) -> None:
        f = fixtures
        repo = f.repo
        repo.files.bulk_save(f.bulk_content_files)
        query = """
          query {
            flSearch(key: "ba*", machine: "polaris") {
              path binpkg { cpvb repo url }
            }
          }
        """
        result = graphql(fixtures.client, query)

        self.assertTrue("errors" not in result, result.get("errors"))
        self.assertEqual(len(result["data"]["flSearch"]), 3)


@given(lib.repo, lib.bulk_content_files, testkit.client)
@where(records_backend="memory", repo="gbp_fl.graphql.queries.repo")
class ResolveQueryCountTests(TestCase):
    query = "query totalFileCount { flCount }"

    query_with_machine = """
      query totalFileCountMachine($machine: String!) {
        flCount(machine: $machine)
      }
    """
    query_with_build = """
      query totalFileCountMachine($machine: String!, $buildId: String!) {
        flCount(machine: $machine, buildId: $buildId)
      }
    """

    def test(self, fixtures: Fixtures) -> None:
        f = fixtures
        repo = f.repo

        repo.files.bulk_save(f.bulk_content_files)
        result = graphql(fixtures.client, self.query)

        self.assertTrue("errors" not in result, result.get("errors"))
        self.assertEqual(result["data"]["flCount"], 6)

    def test_with_no_content_files(self, fixtures: Fixtures) -> None:
        result = graphql(fixtures.client, self.query)

        self.assertTrue("errors" not in result, result.get("errors"))
        self.assertEqual(result["data"]["flCount"], 0)

    def test_with_machine(self, fixtures: Fixtures) -> None:
        f = fixtures
        repo = f.repo

        repo.files.bulk_save(f.bulk_content_files)
        result = graphql(
            fixtures.client, self.query_with_machine, {"machine": "lighthouse"}
        )

        self.assertTrue("errors" not in result, result.get("errors"))
        self.assertEqual(result["data"]["flCount"], 2)

    def test_with_build(self, fixtures: Fixtures) -> None:
        f = fixtures
        repo = f.repo

        repo.files.bulk_save(f.bulk_content_files)
        result = graphql(
            fixtures.client,
            self.query_with_build,
            {"machine": "polaris", "buildId": "26"},
        )

        self.assertTrue("errors" not in result, result.get("errors"))
        self.assertEqual(result["data"]["flCount"], 3)


# Any test that uses "record" depends on Django, because "records" depends on Django.
# This needs to be fixed
@given(testkit.publisher, testkit.record, lib.now)
@where(records_db__backend="django")
class ResolveBinPkgBuildTests(DjangoTestCase):

    def test(self, fixtures: Fixtures) -> None:
        f = fixtures
        build_record: BuildRecord = replace(f.record, submitted=f.now)
        build = Build(machine=build_record.machine, build_id=build_record.build_id)
        binpkg = BinPkg(
            build=build,
            cpvb="dev-language/python-3.13.1-3",
            repo="gentoo",
            build_time=fixtures.now,
        )
        publisher.repo.build_records.save(build_record)
        result = schema.type_map["flBinPkg"].fields["build"].resolve(binpkg, Mock())

        self.assertEqual(result, build_record)


@given(lib.repo, lib.bulk_content_files, testkit.client)
@where(records_backend="memory", repo="gbp_fl.graphql.queries.repo")
class FileListListTests(TestCase):
    def test(self, fixtures: Fixtures) -> None:
        f = fixtures
        repo = f.repo
        repo.files.bulk_save(f.bulk_content_files)

        query = """
          query {
            flList(machine: "lighthouse", buildId: "34", cpvb: "app-shells/bash-5.2_p37-1") {
              path timestamp size
            }
          }
        """
        result = graphql(fixtures.client, query)

        self.assertTrue("errors" not in result, result.get("errors"))
        expected = [
            {
                "path": "/bin/bash",
                "size": 850648,
                "timestamp": "2025-01-26T12:57:37+00:00",
            },
            {
                "path": "/etc/skel",
                "size": 850648,
                "timestamp": "2025-01-26T12:57:37+00:00",
            },
        ]
        self.assertEqual(expected, result["data"]["flList"])


@given(lib.repo, testkit.client, testkit.publisher)
@where(records_backend="memory")
class FlListPackages(TestCase):
    def test(self, fixtures: Fixtures) -> None:
        build = GBPBuild(machine="lighthouse", build_id="34404")
        publisher.publish(build)

        query = """
          query {
            flListPackages(machine: "lighthouse", buildId: "34404") {
              cpvb
              files {
                path timestamp size
              }
            }
          }
        """
        result = graphql(fixtures.client, query)

        self.assertTrue("errors" not in result, result.get("errors"))

        # the files array will be empty since the artifact builder factory doesn't know
        # how to create actual package files.
        expected = [
            {"cpvb": "acct-group/sgx-0-1", "files": []},
            {"cpvb": "app-admin/perl-cleaner-2.30-1", "files": []},
            {"cpvb": "app-arch/unzip-6.0_p26-1", "files": []},
            {"cpvb": "app-crypt/gpgme-1.14.0-1", "files": []},
        ]
        self.assertEqual(expected, result["data"]["flListPackages"])



"""Test utilities"""

# pylint: disable=missing-docstring,redefined-outer-name

import argparse
import datetime as dt
import os
import shlex
from contextlib import contextmanager
from pathlib import PurePath as Path
from tarfile import TarInfo
from typing import Any, Generator, Sequence
from unittest import mock

import gbpcli
from gbp_testkit import fixtures as testkit
from gbp_testkit import helpers
from gbpcli.config import Config
from gbpcli.gbp import GBP
from gbpcli.types import Console
from gentoo_build_publisher import types as gbp
from gentoo_build_publisher import worker as gbp_worker
from unittest_fixtures import FixtureContext, Fixtures, fixture

from gbp_fl.records import Repo
from gbp_fl.settings import Settings
from gbp_fl.types import BinPkg, Build, BuildLike, ContentFile, Package

DEFAULT_CONTENTS = """
    lighthouse 34 app-shells/bash-5.2_p37-1 /bin/bash
    lighthouse 34 app-shells/bash-5.2_p37-1 /etc/skel
    polaris    26 app-arch/tar-1.35-1       /bin/gtar
    polaris    26 app-shells/bash-5.2_p37-1 /bin/bash
    polaris    26 app-shells/bash-5.2_p37-2 /bin/bash
    polaris    27 app-shells/bash-5.2_p37-1 /bin/bash
"""
LOCAL_TIMEZONE = dt.timezone(dt.timedelta(days=-1, seconds=61200), "PDT")
TESTDIR = Path(__file__).parent


class MockGBPGateway:
    """Not a real gateway"""

    def __init__(self) -> None:
        self.builds: list[BuildLike] = []
        self.packages: dict[Build, list[Package]] = {}
        self.contents: dict[tuple[Build, Package], list[TarInfo]] = {}
        self.machines: list[str] = []

    def get_packages(self, build: Build) -> list[Package]:
        try:
            return self.packages[build]
        except KeyError:
            raise LookupError(build) from None

    def get_package_contents(self, build: Build, package: Package) -> list[TarInfo]:
        try:
            return self.contents[build, package]
        except KeyError:
            raise LookupError(build, package) from None

    def list_machine_names(self) -> list[str]:
        return self.machines


@contextmanager
def cd(path: str) -> Generator[None, None, None]:
    cwd = os.getcwd()

    os.chdir(path)
    yield
    os.chdir(cwd)


def parse_args(cmdline: str) -> argparse.Namespace:
    """Return cmdline as parsed arguments"""
    args = shlex.split(cmdline)
    parser = gbpcli.build_parser(Config(url="http://gbp.invalid/"))

    return parser.parse_args(args[1:])


def print_command(cmdline: str, console: Console) -> None:
    """Pretty print the cmdline to console"""
    console.out.print(f"[green]$ [/green]{cmdline}")


@fixture()
def gbp_client(_fixtures: Fixtures, url: str = "http://gbp.invalid/") -> GBP:
    return helpers.test_gbp(url)


@fixture(testkit.tmpdir)
def environ(
    fixtures: Fixtures, environ: dict[str, str] | None = None
) -> FixtureContext[dict[str, str]]:
    mock_environ = {
        **next(testkit.environ(fixtures), {}),
        "BUILD_PUBLISHER_API_KEY_ENABLE": "no",
        "BUILD_PUBLISHER_JENKINS_BASE_URL": "https://jenkins.invalid/",
        "BUILD_PUBLISHER_RECORDS_BACKEND": "memory",
        "BUILD_PUBLISHER_STORAGE_PATH": str(fixtures.tmpdir / "gbp"),
        "BUILD_PUBLISHER_WORKER_BACKEND": "sync",
        "BUILD_PUBLISHER_WORKER_THREAD_WAIT": "yes",
        "GBP_FL_RECORDS_BACKEND": "memory",
        **(environ or {}),
    }
    with mock.patch.dict(os.environ, mock_environ):
        yield mock_environ


@fixture(testkit.tmpdir, environ)
def settings(_fixtures: Fixtures) -> Settings:
    return Settings.from_environ()


@fixture(settings)
def repo(fixtures: Fixtures, repo: str = "gbp_fl.records.repo") -> FixtureContext[Repo]:
    repo_: Repo = Repo.from_settings(fixtures.settings)

    with mock.patch(repo, new=repo_):
        yield repo_


@fixture()
def now(
    _fixtures: Fixtures,
    now: dt.datetime = dt.datetime(2025, 1, 26, 12, 57, 37, tzinfo=dt.UTC),
) -> dt.datetime:
    return now


@fixture()
def build(_fixtures: Fixtures, build: str = "lighthouse.34") -> Build:
    machine, build_id = build.split(".", 1)

    return Build(machine=machine, build_id=build_id)


@fixture(build, now)
def binpkg(  # pylint: disable=too-many-arguments
    fixtures: Fixtures,
    build: Build | None = None,
    cpvb: str = "app-shells/bash-5.2_p37-3",
    build_time: dt.datetime | None = None,
    repo: str = "gentoo",
) -> BinPkg:
    return BinPkg(
        build=build or fixtures.build,
        cpvb=cpvb,
        build_time=build_time or fixtures.now,
        repo=repo,
    )


@fixture(binpkg, now)
def content_file(
    fixtures: Fixtures,
    binpkg: BinPkg | None = None,
    path: Path = Path("/bin/bash"),
    timestamp: dt.datetime | None = None,
    size: int = 870400,
) -> ContentFile:
    return ContentFile(
        binpkg=binpkg or fixtures.binpkg,
        path=path,
        timestamp=timestamp or fixtures.now,
        size=size,
    )


@fixture(now)
def bulk_content_files(
    fixtures: Fixtures, bulk_content_files: str = DEFAULT_CONTENTS
) -> list[ContentFile]:
    content_files: list[ContentFile] = []
    cf_defs: str = bulk_content_files.strip()
    for cf_def in cf_defs.split("\n"):
        cf_def = cf_def.strip()

        if not cf_def:
            continue

        parts = cf_def.split()
        machine, build_id, cpvb, path = parts[:4]

        try:
            repo_ = parts[4]
        except IndexError:
            repo_ = "gentoo"

        try:
            size = int(parts[5])
        except IndexError:
            size = 850648

        try:
            timestamp = dt.datetime.fromisoformat(parts[6]).astimezone(dt.UTC)
        except IndexError:
            timestamp = fixtures.now

        bld = Build(machine=machine, build_id=build_id)
        pkg = BinPkg(build=bld, cpvb=cpvb, build_time=fixtures.now, repo=repo_)
        content_files.append(
            ContentFile(
                binpkg=pkg, path=Path(path), timestamp=timestamp, size=int(size)
            )
        )

    return content_files


@fixture(now)
def bulk_packages(fixtures: Fixtures, bulk_packages: str = "") -> list[Package]:
    packages: list[Package] = []

    for p_def in (bulk_packages or "").strip().split("\n"):
        p_def = p_def.strip()

        if not p_def:
            continue

        parts = p_def.split()
        cpv = parts[0]

        build_id = seq_get(parts, 1, 1)

        try:
            build_time = int(parts[2])
        except IndexError:
            build_time = fixtures.now.timestamp()

        # crude parsing, but good enough for now
        c, pv = cpv.split("/", 1)
        p, v = pv.rsplit("-", 1)
        if v.startswith("r"):
            p, rest = p.rsplit("-", 1)
            v = f"{rest}-{v}"
        path = f"{c}/{p}/{pv}-{build_id}.gpkg.tar"

        package = Package(
            cpv=cpv,
            repo=seq_get(parts, 3, "gentoo"),
            build_id=build_id,
            build_time=build_time,
            path=path,
        )
        packages.append(package)
    return packages


@fixture(testkit.record, now)
def gbp_package(  # pylint: disable=too-many-arguments
    fixtures: Fixtures,
    *,
    build_id: int = 1,
    build_time: int | None = None,
    cpv: str = "sys-libs/mtdev-1.1.7",
    path: str = "sys-libs/mtdev/mtdev-1.1.7-1.gpkg.tar",
    repo: str = "gentoo",
    size: int = 40960,
) -> gbp.Package:
    build_time = build_time or fixtures.now.timestamp()
    return gbp.Package(
        build_id=build_id,
        build_time=build_time,
        cpv=cpv,
        path=path,
        repo=repo,
        size=size,
    )


@fixture(testkit.settings)
def worker(fixtures: Fixtures) -> FixtureContext[gbp_worker.WorkerInterface]:
    sync_worker = gbp_worker.Worker(fixtures.settings)
    with mock.patch("gentoo_build_publisher.worker", sync_worker):
        yield sync_worker


def seq_get(seq: Sequence[Any], index: int, default: Any = None) -> Any:
    """Like dict.get, but for sequences"""
    try:
        return seq[index]
    except IndexError:
        return default


@fixture()
def gateway(_: Fixtures) -> MockGBPGateway:
    return MockGBPGateway()


@fixture()
def local_timezone(
    _: Fixtures, local_timezone: dt.timezone = LOCAL_TIMEZONE
) -> FixtureContext[dt.timezone]:
    with mock.patch("gbpcli.render.LOCAL_TIMEZONE", new=local_timezone):
        yield local_timezone


@fixture()
def tarinfo(_: Fixtures, name: str = "image/bin/bash") -> mock.Mock:
    mock_tarinfo = mock.Mock(mtime=0, size=22)
    mock_tarinfo.isdir.return_value = False
    mock_tarinfo.name = name

    return mock_tarinfo


@fixture()
def package(
    _: Fixtures,
    cpv: str = "sys-libs/mtdev-1.1.7",
    repo: str = "gentoo",
    build_id: int = 1,
    build_time: int = 0,
    path: str = "sys-libs/mtdev/mtdev-1.1.7-1.gpkg.tar",
) -> Package:
    return Package(
        cpv=cpv, repo=repo, build_id=build_id, build_time=build_time, path=path
    )



# pylint: disable=missing-docstring,unused-argument
import logging
import shutil
from pathlib import Path
from unittest import mock

import gbp_testkit.fixtures as testkit
from django.test import TestCase
from unittest_fixtures import Fixtures, fixture, given, where

from gbp_fl import gateway
from gbp_fl.records import Repo

from . import lib

logging.basicConfig(handlers=[logging.NullHandler()])


@given(testkit.publisher)
class GBPTestCase(TestCase):
    options = {"records_backend": "django"}


@fixture(lib.gbp_package, build_record=testkit.record)
def binpkg(f: Fixtures) -> Path:
    gbp = gateway.GBPGateway()
    path = Path(gbp.get_full_package_path(f.build_record, f.gbp_package))
    path.parent.mkdir(parents=True)
    shutil.copy(lib.TESTDIR / "assets/sys-libs/mtdev/mtdev-1.1.7-1.gpkg.tar", path)

    with open(path.parents[2] / "Packages", "w", encoding="utf8") as meta:
        meta.write(PACKAGES)

    return path


# Any test that uses "record" depends on Django, because "records" depends on Django.
# This needs to be fixed
@given(lib.worker, lib.gbp_package, lib.settings, binpkg)
@where(records_db__backend="django")
class PostPulledTests(GBPTestCase):
    def test(self, fixtures: Fixtures) -> None:
        f = fixtures
        repo: Repo = Repo.from_settings(f.settings)
        gbp = gateway.GBPGateway()
        _ = None

        with mock.patch("gbp_fl.package_utils.repo", new=repo):
            gbp.emit_signal(
                "postpull", build=f.build_record, packages=_, gbp_metadata=_
            )

        # pylint: disable=assignment-from-no-return
        content_files = repo.files.for_package(
            f.build_record.machine,
            f.build_record.build_id,
            f"{f.gbp_package.cpv}-{f.gbp_package.build_id}",
        )
        self.assertEqual(len([*content_files]), 10)

    def test_with_empty_tar(self, fixtures: Fixtures) -> None:
        f = fixtures
        repo: Repo = Repo.from_settings(f.settings)
        gbp = gateway.GBPGateway()
        _ = None

        shutil.copy(lib.TESTDIR / "assets/empty.tar", f.binpkg)

        with mock.patch("gbp_fl.package_utils.repo", new=repo):
            gbp.emit_signal(
                "postpull", build=f.build_record, packages=_, gbp_metadata=_
            )

        content_files = list(
            repo.files.for_package(
                f.build_record.machine, f.build_record.build_id, f.gbp_package.cpv
            )
        )
        self.assertEqual(len(content_files), 0)

    def test_pull_with_no_package_manifest(self, fixtures: Fixtures) -> None:
        f = fixtures
        repo: Repo = Repo.from_settings(f.settings)
        gbp = gateway.GBPGateway()
        _ = None

        package_manifest: Path = f.binpkg.parents[2] / "Packages"
        package_manifest.unlink()

        with mock.patch("gbp_fl.package_utils.repo", new=repo):
            gbp.emit_signal(
                "postpull", build=f.build_record, packages=_, gbp_metadata=_
            )

        content_files = list(
            repo.files.for_package(
                f.build_record.machine,
                f.build_record.build_id,
                f"{f.gbp_package.cpv}-{f.gbp_package.build_id}",
            )
        )
        self.assertEqual(len(content_files), 0)


@given(lib.worker, lib.settings, lib.bulk_content_files, lib.build)
@where(build="polaris.26")
class PostDeleteTests(GBPTestCase):
    def test(self, fixtures: Fixtures) -> None:
        f = fixtures
        repo: Repo = Repo.from_settings(f.settings)
        repo.files.bulk_save(f.bulk_content_files)
        gbp = gateway.GBPGateway()

        self.assertEqual(repo.files.count(None, None, None), 6)

        with mock.patch("gbp_fl.records.repo", new=repo):
            gbp.emit_signal("postdelete", build=fixtures.build)

        self.assertEqual(repo.files.count(None, None, None), 3)


# Any test that uses "record" depends on Django, because "records" depends on Django.
# This needs to be fixed
@given(lib.gbp_package, binpkg, build_record=testkit.record)
@where(records_db__backend="django")
class GetPackageContentsTests(GBPTestCase):
    def test(self, fixtures: Fixtures) -> None:
        f = fixtures
        gbp = gateway.GBPGateway()

        contents = gbp.get_package_contents(f.build_record, f.gbp_package)

        self.assertEqual(len(list(contents)), 19)


PACKAGES = """\
ACCEPT_KEYWORDS: amd64 ~amd64
ACCEPT_LICENSE: @FREE
ACCEPT_PROPERTIES: *
ACCEPT_RESTRICT: *
ARCH: amd64
CBUILD: x86_64-pc-linux-gnu
CHOST: x86_64-pc-linux-gnu
ELIBC: glibc
GENTOO_MIRRORS: http://distfiles.gentoo.org
INSTALL_MASK: /usr/share/doc
KERNEL: linux
PACKAGES: 1
TIMESTAMP: 1738033770
VERSION: 0

BUILD_ID: 1
BUILD_TIME: 1725728729
CPV: sys-libs/mtdev-1.1.7
DEFINED_PHASES: configure install
DEPEND: >=sys-kernel/linux-headers-2.6.31
EAPI: 8
KEYWORDS: ~alpha amd64 arm arm64 ~hppa ~ia64 ~loong ~m68k ~mips ppc ppc64 ~riscv ~s390 sparc x86
LICENSE: MIT
MD5: 5c9f95bf9b86a05e7348e14b64badd69
PATH: sys-libs/mtdev/mtdev-1.1.7-1.gpkg.tar
PROVIDES: x86_64: libmtdev.so.1
RDEPEND: >=sys-libs/glibc-2.40
REQUIRES: x86_64: libc.so.6
SHA1: 837d681085179a949527938c71f5f32d8fd8ffc9
SIZE: 40960
USE: abi_x86_64 amd64 elibc_glibc kernel_linux
MTIME: 1725728730
REPO: gentoo
"""



"""Tests for the package_utils module"""

from pathlib import Path
from unittest import TestCase, mock

import gbp_testkit.fixtures as testkit
from django.test import TestCase as DjangoTestCase
from unittest_fixtures import Fixtures, given, where

from gbp_fl import package_utils
from gbp_fl.records import files_backend
from gbp_fl.types import ContentFileInfo

from . import lib

# pylint: disable=missing-docstring,unused-argument

MOCK_PREFIX = "gbp_fl.package_utils."


@given(lib.bulk_packages, lib.gateway, lib.tarinfo, lib.build)
@where(
    bulk_packages="""
    app-crypt/rhash-1.4.5
    dev-libs/libgcrypt-1.11.0-r2
    dev-libs/openssl-3.3.2-r2
    dev-libs/wayland-protocols-1.39
    net-dns/c-ares-1.34.4
"""
)
@where(build="babette.1505")
class IndexBuildTests(TestCase):
    def test(self, fixtures: Fixtures) -> None:
        mock_gw = fixtures.gateway
        package = fixtures.bulk_packages[0]
        build = fixtures.build
        mock_gw.packages[build] = [package]
        mock_gw.contents[build, package] = [fixtures.tarinfo]
        repo = mock.Mock(files=files_backend("memory"))

        with mock.patch(f"{MOCK_PREFIX}gateway", new=mock_gw):
            with mock.patch(f"{MOCK_PREFIX}repo", new=repo):
                package_utils.index_build(build)

        self.assertEqual(repo.files.count(None, None, None), 1)

        content_file = next(iter(repo.files.files.values()))
        self.assertEqual(content_file.path, Path("/bin/bash"))
        self.assertEqual(content_file.size, 22)

    def test_when_no_package(self, fixtures: Fixtures) -> None:
        mock_gw = fixtures.gateway
        repo = mock.Mock(files=files_backend("memory"))

        with mock.patch(f"{MOCK_PREFIX}gateway", new=mock_gw):
            with mock.patch(f"{MOCK_PREFIX}repo", new=repo):
                package_utils.index_build(fixtures.build)

        self.assertEqual(repo.files.count(None, None, None), 0)


# Any test that uses "record" depends on Django, because "records" depends on Django.
# This needs to be fixed
@where(records_db__backend="django")
@given(lib.gbp_package, testkit.record)
class MakeContentFileTests(DjangoTestCase):
    def test(self, fixtures: Fixtures) -> None:
        f = fixtures
        info = ContentFileInfo(name="/bin/bash", mtime=1738258812, size=8829)

        result = package_utils.make_content_file(f.record, f.gbp_package, info)

        self.assertEqual(result.path, Path("/bin/bash"))
        self.assertEqual(result.size, 8829)



"""Handler for `gbp fl fetch`"""

import argparse
from typing import Iterable

import requests
from gbpcli.gbp import GBP
from gbpcli.types import Console
from yarl import URL

from gbp_fl import utils

from . import resolve_build_id

BUFSIZE = 1024
HELP = "Get package files from Gentoo Build Publisher"


def handler(args: argparse.Namespace, gbp: GBP, console: Console) -> int:
    """Get package files from Gentoo Build Publisher"""
    if (spec := utils.parse_pkgspec(args.pkgspec)) is None:
        console.err.print(f"[red]Invalid specifier: {args.pkgspec}[/red]")
        return 1

    build_id = resolve_build_id(spec.machine, spec.build_id, gbp) or ""
    path = (
        f"machines/{spec.machine}/builds/{build_id}/packages/{spec.c}"
        f"/{spec.p}/{spec.p}-{spec.v}-{spec.b}"
    )
    url = URL(gbp.query._url).origin() / path  # pylint: disable=protected-access

    with requests.get(str(url), stream=True, timeout=300) as response:
        if response.status_code == 404:
            console.err.print("[red]The requested package was not found.[/red]")
            return 2

        output = URL(response.url).name

        save_chunks_to_file(response.iter_content(BUFSIZE), output)
        console.out.print(f"package saved as [package]{output}[/package]")

    return 0


def save_chunks_to_file(chunks: Iterable[bytes], filename: str) -> None:
    """Save the content to the given filename"""
    with open(filename, "wb", buffering=BUFSIZE) as fp:
        for chunk in chunks:
            fp.write(chunk)


def parse_args(parser: argparse.ArgumentParser) -> None:
    """Build command-line arguments"""
    parser.add_argument("pkgspec")



"""Gateway to the Gentoo Build Publisher

The rationale here is not to expose Gentoo Build Publisher internals to the rest of
gbp-fl, even though we're using the public interface. I want to do this because I don't
want the rest of gbp-fl importing modules from gentoo_build_publisher. In django there's
this ongoing risk with app dependencies where you might encounter either circular deps
or apps not ready and I want to at least put all that risk in one place.

The GBPGateway here talks to GBP, however each method is responsible for importing what
parts of GBP it needs form *inside* the method itself. In addition methods are required
to only accept gbp-fl-local types and only return gbp-fl-local times. One exception is
the GBP signal receiver functions which will naturally receive gbp-internal types.
However there are wrapper types in gbp_fl.types so that receivers can at least only
access gbp-local attributes or else type checkers will holler.
"""

from contextlib import contextmanager
from pathlib import PurePath as Path
from tarfile import TarFile, TarInfo
from typing import Any, Callable, Iterator, ParamSpec, cast

from pydispatch import Dispatcher  # type: ignore

from gbp_fl import utils
from gbp_fl.types import Build, BuildLike, Package

P = ParamSpec("P")


class GBPGateway:
    """The GBP Gateway

    Methods should only accept gbp-fl types and only return gbp-fl types.
    """

    # pylint: disable=import-outside-toplevel
    def receive_signal(self, receiver: Callable[..., Any], signal: str) -> None:
        """Register the given signal receiver with the given GBP signal

        Receivers may receive GBP data types and are encourage to convert them to gbp-fl
        data types immediately.
        """
        dispatcher = self._dispatcher
        dispatcher.bind(**{signal: receiver})

    def emit_signal(self, signal: str, **kwargs: Any) -> None:
        """Emit signal on the gbp dispatcher

        This is mainly used for testing
        """
        dispatcher = self._dispatcher
        dispatcher.emit(signal, **kwargs)

    def list_machine_names(self) -> list[str]:
        """Return the list of machines that GBP holds builds for"""
        from gentoo_build_publisher import publisher

        repo = publisher.repo

        return repo.build_records.list_machines()

    def get_builds_for_machine(self, machine: str) -> Iterator[Build]:
        """Return the builds for the given machine"""
        from gentoo_build_publisher import publisher

        repo = publisher.repo

        return (
            Build(machine=build.machine, build_id=build.build_id)
            for build in repo.build_records.for_machine(machine)
        )

    def get_build_record(self, build: Build) -> BuildLike:
        """Return the build record given the build

        Internally we should only treat this like a BuildLike, however GraphQL may treat
        it like a GBP BuildRecord.
        """
        from gentoo_build_publisher import publisher
        from gentoo_build_publisher.types import Build as GBPBuild

        build_records = publisher.repo.build_records

        record = build_records.get(
            GBPBuild(machine=build.machine, build_id=build.build_id)
        )

        # I dont' think you should need to cast but mypy begs to differ
        return cast(BuildLike, record)

    def get_full_package_path(self, build: Build, package: Package) -> Path:
        """Return the full path of the given Package"""
        from gentoo_build_publisher import publisher, types

        storage = publisher.storage
        b = types.Build(machine=build.machine, build_id=build.build_id)
        binpkgs_path = storage.get_path(b, types.Content.BINPKGS)

        return binpkgs_path / package.path

    def get_packages(self, build: Build) -> list[Package]:
        """Return all the Packages contained in the given Build"""
        from gentoo_build_publisher import publisher, types

        storage = publisher.storage
        gbp_build = types.Build(machine=build.machine, build_id=build.build_id)

        return [
            Package(
                cpv=p.cpv,
                repo=p.repo,
                path=p.path,
                build_id=p.build_id,
                build_time=p.build_time,
            )
            for p in storage.get_packages(gbp_build)
        ]

    def get_package_contents(self, build: Build, package: Package) -> Iterator[TarInfo]:
        """Given the build and binary package, return the packages contents

        This scours the binary tarball for package files.
        Generates tarfile.TarInfo objects.
        """
        package_path = self.get_full_package_path(build, package)

        with TarFile.open(package_path, "r") as tarfile:
            utils.ensure_package_identifier(package, tarfile)

            # We're not sure of the exact filename of the inner tarfile because of
            # available compression options, but we know what the name starts with.
            # https://www.gentoo.org/glep/glep-0078.html#the-container-format
            pv = package.cpv.partition("/")[2]
            prefix = f"{pv}-{package.build_id}/image.tar"

            for item in tarfile:  # pragma: no branch
                if item.name.startswith(prefix):
                    image_fp = tarfile.extractfile(item)
                    # this is also a tarfile
                    with TarFile.open(mode="r", fileobj=image_fp) as image:
                        yield from image.getmembers()
                    break

    def run_task(
        self, func: Callable[P, Any], *args: P.args, **kwargs: P.kwargs
    ) -> None:
        """Send the given callable (and args to the GBP task worker to run"""
        from gentoo_build_publisher import worker

        worker.run(func, *args, **kwargs)

    def _really_set_process(self, build: Build, phase: str) -> None:
        """Unconditionally set the current process state for the given build"""
        from gbp_ps.signals import set_process
        from gentoo_build_publisher import types

        gbp_build = types.Build(machine=build.machine, build_id=build.build_id)
        set_process(gbp_build, phase)

    @contextmanager
    def set_process(self, build: Build, phase: str) -> Iterator[bool]:
        """Conditinally set the current process state for the given build

        When entered, sets the process to the given phase, when exited, set's the
        process phase to "clean".

        If the gbp-ps plugin is not installed, this is a noop.

        Yield True if the process was set, otherwise yield False.
        """
        if self.has_plugin("gbp-ps"):
            self._really_set_process(build, phase)
            yield True
            self._really_set_process(build, "clean")
        else:
            yield False

    @staticmethod
    def has_plugin(name: str) -> bool:
        """Return true if gbp has the given plugin"""
        from gentoo_build_publisher import plugins

        return any(plugin.name == name for plugin in plugins.get_plugins())

    @property
    def _dispatcher(self) -> Dispatcher:
        """Return the GBP signal dispatcher.

        Warning: this method leaks!
        """
        from gentoo_build_publisher import signals

        return signals.dispatcher


gateway = GBPGateway()



"""Utilities for gbp-fl"""

import re
from dataclasses import dataclass
from tarfile import TarFile

from gbp_fl.types import MissingPackageIdentifier, Package

# firefox-135.0-1
PKGSPEC_RE_STR = r"""
(?P<p>[a-z].*)-
(?P<v>[0-9].*)-
(?P<b>[0-9]*)
"""

PKGSPEC_RE = re.compile(PKGSPEC_RE_STR, re.I | re.X)


@dataclass
class Parsed:
    """Parsed package spec"""

    machine: str
    build_id: str
    c: str
    p: str
    v: str
    b: int

    @property
    def cpvb(self) -> str:
        """The cpvb (category, version, package, build_id) for the package"""
        return f"{self.c}/{self.p}-{self.v}-{self.b}"


def parse_pkgspec(pkgspec: str) -> Parsed | None:
    """Parse the given spec"""
    parts = pkgspec.split("/")

    if len(parts) != 4:
        return None

    machine, build_id, c, pvb = parts

    if match := PKGSPEC_RE.match(pvb):
        p, v, b = match.groups()
        parsed = Parsed(machine=machine, build_id=build_id, c=c, p=p, v=v, b=int(b))
        return parsed
    return None


def ensure_package_identifier(package: Package, tarfile: TarFile) -> None:
    """Raise LookupError if tarfile is missing the required identifier"""
    pv = package.cpv.partition("/")[2]
    package_identifier = f"{pv}-{package.build_id}/gpkg-1"

    if package_identifier in tarfile.getnames():
        return

    msg = f"Expected {package_identifier} in the archive, but it was not found."
    raise MissingPackageIdentifier(msg)



"""gbp-fl

A pfl-inspired plugin for Gentoo Build Publisher.
"""

import importlib.metadata

__version__ = importlib.metadata.version("gbp-fl")

# Plugin definition
plugin = {
    "name": "gbp-fl",
    "app": "gbp_fl.django.gbp_fl.apps.GBPFLConfig",
    "description": "List and search build's package files",
    "version": __version__,
    "graphql": "gbp_fl.graphql",
}



"""The Query GraphQL type for gbp-fl"""

import datetime as dt
from functools import partial
from typing import Any, TypeAlias

from ariadne import ObjectType, convert_kwargs_to_snake_case
from graphql import GraphQLResolveInfo

from gbp_fl.gateway import gateway
from gbp_fl.records import repo
from gbp_fl.types import BinPkg, Build, ContentFile

Info: TypeAlias = GraphQLResolveInfo
Query = ObjectType("Query")

# pylint: disable=missing-docstring


@Query.field("flSearch")
def _(
    _obj: Any, _info: Info, *, key: str, machine: str | None = None
) -> list[ContentFile]:
    return list(repo.files.search(key, [machine] if machine else None))


@Query.field("flSearchV2")
def _(
    _obj: Any, _info: Info, *, key: str, machines: list[str] | None = None
) -> list[ContentFile]:
    return list(repo.files.search(key, machines))


@Query.field("flCount")
@convert_kwargs_to_snake_case
def _(
    _obj: Any, _info: Info, *, machine: str | None = None, build_id: str | None = None
) -> int:
    return repo.files.count(machine, build_id, None)


@Query.field("flList")
@convert_kwargs_to_snake_case
def _(
    _obj: Any, _info: Info, *, machine: str, build_id: str, cpvb: str
) -> list[ContentFile]:
    return list(repo.files.for_package(machine, build_id, cpvb))


@Query.field("flListPackages")
@convert_kwargs_to_snake_case
def _(_obj: Any, _info: Info, *, machine: str, build_id: str) -> list[BinPkg]:
    build = Build(machine=machine, build_id=build_id)
    time = partial(dt.datetime.fromtimestamp, tz=dt.UTC)

    return [
        BinPkg(build=build, cpvb=p.cpvb, repo=p.repo, build_time=time(p.build_time))
        for p in gateway.get_packages(build)
    ]



"""Tests for gbp-fl async tasks"""

# The tasks, by design, do basically nothing. We just have to assert the call the
# appropriate functions with the appropriate args

from unittest import TestCase, mock

from unittest_fixtures import Fixtures, given

from gbp_fl.worker import tasks

from . import lib

# pylint: disable=missing-docstring


@given(lib.build)
class IndexBuildTests(TestCase):
    @mock.patch("gbp_fl.gateway.GBPGateway.set_process")
    @mock.patch("gbp_fl.package_utils")
    def test(
        self, package_utils: mock.Mock, set_process: mock.Mock, fixtures: Fixtures
    ) -> None:
        build = fixtures.build
        tasks.index_build(build.machine, build.build_id)

        package_utils.index_build(build)

        set_process.assert_called_once_with(build, "index")


@given(lib.build)
class DeindexBuildTests(TestCase):
    @mock.patch("gbp_fl.gateway.GBPGateway.set_process")
    @mock.patch("gbp_fl.records.Repo.from_settings")
    def test(
        self, repo_from_settings: mock.Mock, set_process: mock.Mock, fixtures: Fixtures
    ) -> None:
        build = fixtures.build
        tasks.deindex_build(build.machine, build.build_id)

        repo = repo_from_settings.return_value
        repo.files.deindex_build.assert_called_once_with(build.machine, build.build_id)

        set_process.assert_called_once_with(build, "deindex")



"""Utilities for working with Packages"""

import datetime as dt
from concurrent.futures import ThreadPoolExecutor, wait
from functools import partial
from pathlib import PurePath as Path

from gbp_fl.gateway import gateway
from gbp_fl.records import repo
from gbp_fl.types import BinPkg, Build, ContentFile, ContentFileInfo, Package


def index_build(build: Build) -> None:
    """Save the given Build's packages to the database"""
    executor = ThreadPoolExecutor()

    try:
        packages = gateway.get_packages(build) or []
    except LookupError:
        return

    wait(executor.submit(index_package, package, build) for package in packages)


def index_package(package: Package, build: Build) -> None:
    """Save the files from the given build/package"""
    items = gateway.get_package_contents(build, package)
    content_file = partial(make_content_file, build, package)

    repo.files.bulk_save(
        content_file(ContentFileInfo(name=i.name, mtime=int(i.mtime), size=i.size))
        for i in items
        if not i.isdir() and i.name.startswith("image/")
    )


def make_content_file(
    build: Build, gbp_package: Package, metadata: ContentFileInfo
) -> ContentFile:
    """Return a ContentFile given the parameters"""
    return ContentFile(
        path=Path(metadata.name.removeprefix("image")),
        binpkg=make_binpkg(build, gbp_package),
        size=metadata.size,
        timestamp=dt.datetime.fromtimestamp(metadata.mtime, tz=dt.UTC),
    )


def make_binpkg(build: Build, gbp_package: Package) -> BinPkg:
    """Create a BinPkg given the build and Package"""
    return BinPkg(
        build=Build(machine=build.machine, build_id=build.build_id),
        cpvb=f"{gbp_package.cpv}-{gbp_package.build_id}",
        repo=gbp_package.repo,
        build_time=dt.datetime.fromtimestamp(gbp_package.build_time, dt.UTC),
    )



"""Tests for gbp_fl.utils"""

from unittest import TestCase

from gbp_fl.utils import Parsed, parse_pkgspec

# pylint: disable=missing-docstring


class ParsePkgspecTests(TestCase):
    def test_host_with_dash(self) -> None:
        parsed = parse_pkgspec("jenkins-python/211/dev-python/anyio-4.8.0-1")

        self.assertIsNotNone(parsed)

        expected = Parsed(
            machine="jenkins-python",
            build_id="211",
            c="dev-python",
            p="anyio",
            v="4.8.0",
            b=1,
        )
        self.assertEqual(expected, parsed)

    def test_invalid_pvb(self) -> None:
        self.assertIsNone(parse_pkgspec("jenkins-python/211/dev-python/?"))



# pylint: disable=missing-docstring,unused-argument
from unittest import TestCase

import gbp_testkit.fixtures as testkit
from gentoo_build_publisher import publisher
from gentoo_build_publisher.types import Build as GBPBuild
from unittest_fixtures import Fixtures, given, where

from gbp_fl.cli import ls

from . import lib

BULK_CONTENT_FILES = """
lighthouse 34 app-arch/tar-1.35-1 /usr/share/info/tar.info-2        gentoo  317618 2025-02-08T07:34:00
lighthouse 34 app-arch/tar-1.35-1 /usr/share/info/tar.info-3        gentoo   49627 2025-02-08T07:34:00
lighthouse 34 app-arch/tar-1.35-1 /usr/share/man/man1/gtar.1        gentoo   42162 2025-02-08T07:34:00
lighthouse 34 app-arch/tar-1.35-1 /usr/share/man/man8/grmt.8        gentoo    5361 2025-02-08T07:34:00
lighthouse 34 app-arch/tar-1.35-1 /usr/share/doc/tar-1.35/README    gentoo    9756 2025-02-08T07:34:00
lighthouse 34 app-arch/tar-1.35-1 /usr/share/doc/tar-1.35/ChangeLog gentoo  579457 2025-02-08T07:34:00
lighthouse 34 app-arch/tar-1.35-1 /usr/share/doc/tar-1.35/AUTHORS   gentoo     601 2025-02-08T07:34:00
lighthouse 34 app-arch/tar-1.35-1 /usr/share/info/tar.info          gentoo   13479 2025-02-08T07:34:00
lighthouse 34 app-arch/tar-1.35-1 /usr/share/info/tar.info-1        gentoo  304308 2025-02-08T07:34:00
lighthouse 34 app-arch/tar-1.35-1 /usr/share/man/man8/rmt.8         gentoo       0 2025-02-08T07:34:01
lighthouse 34 app-arch/tar-1.35-1 /usr/share/doc/tar-1.35/NEWS      gentoo   67627 2025-02-08T07:34:01
lighthouse 34 app-arch/tar-1.35-1 /usr/share/doc/tar-1.35/THANKS    gentoo   20500 2025-02-08T07:34:01
lighthouse 34 app-arch/tar-1.35-1 /usr/share/doc/tar-1.35/TODO      gentoo    2151 2025-02-08T07:34:01
lighthouse 34 app-arch/tar-1.35-1 /bin/gtar                         gentoo  411312 2025-02-08T07:34:01
polaris    26 app-arch/tar-1.35-1       /bin/gtar
polaris    26 app-shells/bash-5.2_p37-1 /bin/bash
polaris    26 app-shells/bash-5.2_p37-2 /bin/bash
polaris    27 app-shells/bash-5.2_p37-1 /bin/bash
"""


@given(
    lib.local_timezone,
    lib.bulk_content_files,
    testkit.console,
    lib.gbp_client,
    lib.repo,
    testkit.publisher,
)
@where(
    records_db={"records_backend": "memory"},
    bulk_content_files=BULK_CONTENT_FILES,
    repo="gbp_fl.graphql.queries.repo",
)
class LsTests(TestCase):

    def test_short_format(self, fixtures: Fixtures) -> None:
        cfs = fixtures.bulk_content_files
        repo = fixtures.repo
        repo.files.bulk_save(cfs)

        pkgspec = "lighthouse/34/app-arch/tar-1.35-1"
        cmd = f"gbp fl ls {pkgspec}"
        args = lib.parse_args(cmd)
        console = fixtures.console
        gbp = fixtures.gbp_client

        lib.print_command(cmd, console)
        status = ls.handler(args, gbp, console)

        self.assertEqual(0, status)
        self.assertEqual(LS_OUTPUT, console.out.file.getvalue())

    def test_long_format(self, fixtures: Fixtures) -> None:
        cfs = fixtures.bulk_content_files
        repo = fixtures.repo
        repo.files.bulk_save(cfs)

        pkgspec = "lighthouse/34/app-arch/tar-1.35-1"
        cmd = f"gbp fl ls -l {pkgspec}"
        args = lib.parse_args(cmd)
        console = fixtures.console
        gbp = fixtures.gbp_client

        lib.print_command(cmd, console)
        status = ls.handler(args, gbp, console)

        self.assertEqual(0, status)
        self.assertEqual(LS_LONG_OUTPUT, console.out.file.getvalue())

    def test_with_tag(self, fixtures: Fixtures) -> None:
        cfs = fixtures.bulk_content_files
        repo = fixtures.repo
        repo.files.bulk_save(cfs)

        publisher.publish(GBPBuild(machine="lighthouse", build_id="34"))
        pkgspec = "lighthouse/@/app-arch/tar-1.35-1"
        cmd = f"gbp fl ls -l {pkgspec}"
        args = lib.parse_args(cmd)
        console = fixtures.console
        gbp = fixtures.gbp_client

        lib.print_command(cmd, console)
        status = ls.handler(args, gbp, console)

        self.assertEqual(0, status)
        expected = LS_LONG_OUTPUT.replace("lighthouse/34/", "lighthouse/@/")
        self.assertEqual(expected, console.out.file.getvalue())

    def test_invalid_spec(self, fixtures: Fixtures) -> None:
        pkgspec = "lighthouse/34/bash-5.2_p37-1"
        cmd = f"gbp fl ls {pkgspec}"
        args = lib.parse_args(cmd)
        gbp = fixtures.gbp_client
        console = fixtures.console

        status = ls.handler(args, gbp, console)

        self.assertEqual(status, 1)
        self.assertEqual(f"Invalid specifier: {pkgspec}\n", console.err.file.getvalue())
        self.assertEqual("", console.out.file.getvalue())

    def test_package_doesnt_exist(self, fixtures: Fixtures) -> None:
        pkgspec = "lighthouse/34/sys-apps/bogus-0.0-1"
        cmd = f"gbp fl ls {pkgspec}"
        args = lib.parse_args(cmd)
        gbp = fixtures.gbp_client
        console = fixtures.console

        status = ls.handler(args, gbp, console)

        self.assertEqual(status, 0)
        self.assertEqual("", console.out.file.getvalue())


LS_OUTPUT = """$ gbp fl ls lighthouse/34/app-arch/tar-1.35-1
/bin/gtar
/usr/share/doc/tar-1.35/AUTHORS
/usr/share/doc/tar-1.35/ChangeLog
/usr/share/doc/tar-1.35/NEWS
/usr/share/doc/tar-1.35/README
/usr/share/doc/tar-1.35/THANKS
/usr/share/doc/tar-1.35/TODO
/usr/share/info/tar.info
/usr/share/info/tar.info-1
/usr/share/info/tar.info-2
/usr/share/info/tar.info-3
/usr/share/man/man1/gtar.1
/usr/share/man/man8/grmt.8
/usr/share/man/man8/rmt.8
"""
LS_LONG_OUTPUT = """$ gbp fl ls -l lighthouse/34/app-arch/tar-1.35-1
╭────────┬───────────────────┬───────────────────────────────────╮
│   Size │ Timestamp         │ Path                              │
├────────┼───────────────────┼───────────────────────────────────┤
│ 411312 │ 02/08/25 06:34:01 │ /bin/gtar                         │
│    601 │ 02/08/25 06:34:00 │ /usr/share/doc/tar-1.35/AUTHORS   │
│ 579457 │ 02/08/25 06:34:00 │ /usr/share/doc/tar-1.35/ChangeLog │
│  67627 │ 02/08/25 06:34:01 │ /usr/share/doc/tar-1.35/NEWS      │
│   9756 │ 02/08/25 06:34:00 │ /usr/share/doc/tar-1.35/README    │
│  20500 │ 02/08/25 06:34:01 │ /usr/share/doc/tar-1.35/THANKS    │
│   2151 │ 02/08/25 06:34:01 │ /usr/share/doc/tar-1.35/TODO      │
│  13479 │ 02/08/25 06:34:00 │ /usr/share/info/tar.info          │
│ 304308 │ 02/08/25 06:34:00 │ /usr/share/info/tar.info-1        │
│ 317618 │ 02/08/25 06:34:00 │ /usr/share/info/tar.info-2        │
│  49627 │ 02/08/25 06:34:00 │ /usr/share/info/tar.info-3        │
│  42162 │ 02/08/25 06:34:00 │ /usr/share/man/man1/gtar.1        │
│   5361 │ 02/08/25 06:34:00 │ /usr/share/man/man8/grmt.8        │
│      0 │ 02/08/25 06:34:01 │ /usr/share/man/man8/rmt.8         │
╰────────┴───────────────────┴───────────────────────────────────╯
"""



"""Async tasks for gbp-fl"""

# pylint: disable=import-outside-toplevel


def index_build(machine: str, build_id: str) -> None:
    """Index packages for the given build"""
    import logging

    from gbp_fl import package_utils
    from gbp_fl.gateway import gateway
    from gbp_fl.types import Build

    logger = logging.getLogger(__package__)
    build = Build(machine=machine, build_id=build_id)

    logger.info("Saving packages for %s.%s", machine, build_id)
    with gateway.set_process(build, "index"):
        package_utils.index_build(build)


def deindex_build(machine: str, build_id: str) -> None:
    """Delete all the files from the given build"""
    from gbp_fl.gateway import gateway
    from gbp_fl.records import repo
    from gbp_fl.types import Build

    files = repo.files
    build = Build(machine=machine, build_id=build_id)

    with gateway.set_process(build, "deindex"):
        files.deindex_build(machine, build_id)



"""Tests for gbp_fl.types"""

# pylint: disable=missing-docstring
from unittest import TestCase

from unittest_fixtures import Fixtures, given, where

from gbp_fl.types import Package

from . import lib


@given(lib.now, lib.binpkg)
@where(binpkg__cpvb="x11-apps/xhost-1.0.10-3")
class BinPkgTests(TestCase):
    def test_cpv(self, fixtures: Fixtures) -> None:
        self.assertEqual(fixtures.binpkg.cpv, "x11-apps/xhost-1.0.10")

    def test_build_id(self, fixtures: Fixtures) -> None:
        self.assertEqual(fixtures.binpkg.build_id, 3)


class PackageTests(TestCase):
    def test_cpvb(self) -> None:
        p = Package(
            cpv="x11-apps/xhost-1.0.10",
            build_id=3,
            repo="gentoo",
            build_time=0,
            path="x11-apps/xhost/xhost-1.0.10-3.gpkg.tar",
        )

        self.assertEqual("x11-apps/xhost-1.0.10-3", p.cpvb)



"""Tests for the gateway interface"""

# pylint: disable=missing-docstring,unused-argument

from contextlib import ExitStack
from pathlib import PurePath as Path
from unittest import TestCase, mock

import gentoo_build_publisher
from gentoo_build_publisher import types as gtype
from unittest_fixtures import FixtureContext, Fixtures, given

from gbp_fl import gateway as gw
from gbp_fl.types import Build, MissingPackageIdentifier

from . import lib


def mock_publisher(_f: Fixtures) -> FixtureContext[dict[str, mock.Mock]]:
    mocks = {"storage": mock.Mock(), "jenkins": mock.Mock(), "repo": mock.Mock()}
    contexts = (
        mock.patch.object(gentoo_build_publisher.publisher, name, value)
        for name, value in mocks.items()
    )

    with ExitStack() as stack:
        for cm in contexts:
            stack.enter_context(cm)

        yield mocks


@given(mock_publisher, lib.build, lib.package)
class GetFullPackagePathTests(TestCase):
    def test(self, fixtures: Fixtures) -> None:
        build = fixtures.build
        mocks = fixtures.mock_publisher
        storage = mocks["storage"]
        build_str = f"{build.machine}.{build.build_id}"

        storage.get_path.return_value = Path(f"/binpkgs/{build_str}")
        gbp = gw.GBPGateway()

        full_package_path = gbp.get_full_package_path(build, fixtures.package)

        self.assertEqual(
            str(full_package_path),
            f"/binpkgs/{build_str}/sys-libs/mtdev/mtdev-1.1.7-1.gpkg.tar",
        )


@given(mock_publisher, lib.build, lib.package)
class GetPackagesTests(TestCase):
    def test(self, fixtures: Fixtures) -> None:
        mocks = fixtures.mock_publisher
        storage = mocks["storage"]
        package = fixtures.package

        storage.get_packages.return_value = [
            gtype.Package(
                cpv=package.cpv,
                build_id=package.build_id,
                repo=package.repo,
                path=package.path,
                size=200,
                build_time=package.build_time,
            )
        ]
        gbp = gw.GBPGateway()

        packages = gbp.get_packages(fixtures.build)

        expected = [package]
        self.assertEqual(packages, expected)


@given(mock_publisher, lib.build, lib.package)
class GetPackageContentsTests(TestCase):
    def test(self, fixtures: Fixtures) -> None:
        mocks = fixtures.mock_publisher
        storage = mocks["storage"]
        storage.get_path.return_value = lib.TESTDIR / "assets"

        gbp = gw.GBPGateway()
        result = list(gbp.get_package_contents(fixtures.build, fixtures.package))

        self.assertEqual(len(result), 19)

    def test_when_empty_tarfile(self, fixtures: Fixtures) -> None:
        gbp = gw.GBPGateway()

        with mock.patch.object(
            gbp, "get_full_package_path", return_value=lib.TESTDIR / "assets/empty.tar"
        ):
            with self.assertRaises(MissingPackageIdentifier):
                list(gbp.get_package_contents(fixtures.build, fixtures.package))


class ReceiveSignalTests(TestCase):
    def test(self) -> None:
        dispatcher = mock.Mock()
        receiver = mock.Mock()
        signal = "sos"

        gbp = gw.GBPGateway()

        with mock.patch("gentoo_build_publisher.signals.dispatcher", dispatcher):
            gbp.receive_signal(receiver, signal)

        dispatcher.bind.assert_called_once_with(sos=receiver)


class EmitSignalTests(TestCase):
    def test(self) -> None:
        dispatcher = mock.Mock()
        signal = "sos"

        gbp = gw.GBPGateway()

        with mock.patch("gentoo_build_publisher.signals.dispatcher", dispatcher):
            gbp.emit_signal(signal, buckle_my="shoe")

        dispatcher.emit.assert_called_once_with(signal, buckle_my="shoe")


class RunTaskTests(TestCase):
    def test(self) -> None:
        gbp = gw.GBPGateway()
        func = mock.Mock()

        with mock.patch("gentoo_build_publisher.worker") as worker:
            gbp.run_task(func, 1, 2, buckle_my="shoe")

        worker.run.assert_called_once_with(func, 1, 2, buckle_my="shoe")


class ListMachineNamesTests(TestCase):
    @mock.patch("gentoo_build_publisher.publisher")
    def test(self, publisher: mock.Mock) -> None:
        machines = ["babette", "lighthouse", "polaris"]
        publisher.repo.build_records.list_machines.return_value = machines

        gbp = gw.GBPGateway()

        self.assertEqual(gbp.list_machine_names(), machines)
        publisher.repo.build_records.list_machines.assert_called_once_with()


class GetBuildsForMachineTests(TestCase):
    @mock.patch("gentoo_build_publisher.publisher")
    def test(self, publisher: mock.Mock) -> None:
        builds = [Build(machine="babette", build_id=f"150{i}") for i in range(5)]
        publisher.repo.build_records.for_machine.return_value = iter(builds)

        gbp = gw.GBPGateway()

        self.assertEqual(list(gbp.get_builds_for_machine("babette")), builds)
        publisher.repo.build_records.for_machine.assert_called_once_with("babette")


@given(lib.build)
class GetBuildRecordTests(TestCase):
    @mock.patch("gentoo_build_publisher.publisher")
    def test(self, publisher: mock.Mock, fixtures: Fixtures) -> None:
        build = fixtures.build
        bdict = {"machine": build.machine, "build_id": build.build_id}
        build_record = mock.Mock(**bdict)
        publisher.repo.build_records.get.return_value = build_record
        gbp = gw.GBPGateway()

        result = gbp.get_build_record(Build(**bdict))

        self.assertEqual(result, build_record)
        publisher.repo.build_records.get.assert_called_once_with(gtype.Build(**bdict))


@given(lib.build)
class SetProcessTests(TestCase):
    @mock.patch("gbp_ps.signals.set_process")
    def test(self, set_process: mock.Mock, fixtures: Fixtures) -> None:
        build = fixtures.build
        gbp = gw.GBPGateway()

        with gbp.set_process(build, "index") as was_set:
            self.assertTrue(was_set)

        gbuild = gtype.Build(machine=build.machine, build_id=build.build_id)
        set_process.assert_has_calls(
            [mock.call(gbuild, "index"), mock.call(gbuild, "clean")]
        )

    @mock.patch("gbp_ps.signals.set_process")
    def test_when_no_gbp_ps_plugin(
        self, set_process: mock.Mock, fixtures: Fixtures
    ) -> None:
        gbp = gw.GBPGateway()

        with mock.patch.object(gbp, "has_plugin", return_value=False):
            with gbp.set_process(fixtures.build, "index") as was_set:
                self.assertFalse(was_set)

        set_process.assert_not_called()


class HasPluginTests(TestCase):
    def test_true(self) -> None:
        self.assertTrue(gw.GBPGateway.has_plugin("gbp-fl"))

    def test_false(self) -> None:
        self.assertFalse(gw.GBPGateway.has_plugin("bogus"))



"""gbp-fl signal handlers"""

from typing import Any, Callable, TypeAlias

from gbp_fl.gateway import gateway
from gbp_fl.types import BuildLike
from gbp_fl.worker import tasks

Receiver: TypeAlias = Callable[..., Any]


def gbp_build_pulled(*, build: BuildLike, **kwargs: Any) -> None:
    """Save the pulled build's package files to the database"""
    gateway.run_task(tasks.index_build, build.machine, build.build_id)


def gbp_build_deleted(*, build: BuildLike) -> None:
    """Signal handler for the deleting of builds

    Delete all the ContentFiles associated with the given build.
    """
    gateway.run_task(tasks.deindex_build, build.machine, build.build_id)


def init() -> None:
    """Initialize"""
    gateway.receive_signal(gbp_build_pulled, "postpull")
    gateway.receive_signal(gbp_build_deleted, "postdelete")



"""GraphQL interface for gbp-fl"""

from importlib import resources

from ariadne import gql

from .binpkg import flBinPkg
from .content_file import flContentFile
from .mutations import Mutation
from .queries import Query

type_defs = gql(resources.read_text("gbp_fl.graphql", "schema.graphql"))
resolvers = [flBinPkg, flContentFile, Query, Mutation]



"""The GraphQL BinPkg type for gbp-fl"""

import re
from typing import TypeAlias

from ariadne import ObjectType
from django.http import HttpRequest
from django.urls import reverse
from graphql import GraphQLResolveInfo

from gbp_fl.gateway import gateway
from gbp_fl.records import Repo
from gbp_fl.settings import Settings
from gbp_fl.types import BinPkg, Build, BuildLike, ContentFile

flBinPkg = ObjectType("flBinPkg")
Info: TypeAlias = GraphQLResolveInfo


# Version regex for cpv's
V_RE = re.compile("-[0-9]")

# pylint: disable=missing-docstring


@flBinPkg.field("build")
def _(pkg: BinPkg, _info: Info) -> BuildLike:
    build = pkg.build

    return gateway.get_build_record(
        Build(machine=build.machine, build_id=build.build_id)
    )


@flBinPkg.field("url")
def _(pkg: BinPkg, info: Info) -> str | None:
    request: HttpRequest = info.context["request"]
    c, pv = pkg.cpv.split("/", 1)

    v_match = V_RE.search(pv)
    assert v_match
    build = pkg.build
    view_args = {
        "machine": build.machine,
        "build_id": build.build_id,
        "c": c,
        "p": pv[: v_match.start()],
        "pv": pv,
        "b": pkg.build_id,
    }
    return request.build_absolute_uri(reverse("gbp-binpkg", kwargs=view_args))


@flBinPkg.field("files")
def _(pkg: BinPkg, _info: Info) -> list[ContentFile]:
    repo = Repo.from_settings(Settings.from_environ())
    return list(repo.files.for_package(pkg.build.machine, pkg.build.build_id, pkg.cpvb))



"""gbp-fl data types

Builds contain BinPkgs and BinPkgs contain ContentFiles, which have Paths and other
metadata.
"""

import datetime as dt
from dataclasses import dataclass
from pathlib import PurePath as Path
from typing import Protocol


@dataclass(frozen=True)
class Package:
    """GBP Package proxy object"""

    cpv: str
    repo: str
    build_id: int
    build_time: int
    path: str

    @property
    def cpvb(self) -> str:
        """The cpv-b string for the Package"""
        return f"{self.cpv}-{self.build_id}"


@dataclass(frozen=True, kw_only=True, slots=True)
class Build:
    """A GBP Build"""

    machine: str
    build_id: str


@dataclass(frozen=True, kw_only=True, slots=True)
class BinPkg:
    """A (binary) package"""

    build: Build

    cpvb: str
    """category-package-version-build_id"""

    repo: str
    """Repository where the package was built from"""

    build_time: dt.datetime

    @property
    def cpv(self) -> str:
        """The BinPkg's cpv"""
        return self.cpvb.rsplit("-", 1)[0]

    @property
    def build_id(self) -> int:
        """The BinPkg's build id"""
        return int(self.cpvb.rsplit("-", 1)[1])


@dataclass(frozen=True, kw_only=True, slots=True)
class ContentFile:
    """A file in a BinPkg (in a Build)"""

    binpkg: BinPkg
    path: Path
    timestamp: dt.datetime

    size: int
    """size of file in bytes"""


class BuildLike(Protocol):  # pylint: disable=too-few-public-methods
    """A GBP Build that we want to pretend we don't know is a gbp-fl Build"""

    machine: str
    build_id: str


@dataclass(frozen=True)
class ContentFileInfo:
    """Interface for ContentFile metadata

    IRL this wrapper around the tarfile.TarInfo object
    """

    # pylint: disable=too-few-public-methods, missing-docstring
    name: str
    """name of the image file"""

    mtime: int
    """modification time in seconds since epoch"""

    size: int
    """file size in bytes"""


class MissingPackageIdentifier(LookupError):
    """Raised when tar archive is missing GLEP-78 package identifier"""



"""search subcommand for gbp-fl"""

import argparse
import datetime as dt
from types import SimpleNamespace
from typing import Any

from gbpcli import render, utils
from gbpcli.gbp import GBP
from gbpcli.types import Console
from rich import box
from rich.table import Table

HELP = "Search for files in packages"


def handler(args: argparse.Namespace, gbp: GBP, console: Console) -> int:
    """Search for files in packages"""
    machines: list[str] | None

    if args.machine:
        machines = [args.machine]
    elif args.mine:
        machines = utils.get_my_machines_from_args(args)
    else:
        machines = None

    data, _ = gbp.query.gbp_fl.searchv2(key=args.key, machines=machines)  # type: ignore

    if content_files := data["flSearchV2"]:
        table = create_table()
        row = table.add_row

        for item in content_files:
            if item["binpkg"]["build"]:
                row(*format_content_file(item, args))
        console.out.print(table)
    return 0


def parse_args(parser: argparse.ArgumentParser) -> None:
    """Set subcommand arguments"""
    parser.add_argument(
        "--machine", "-m", default=None, help="Restrict search to the given machine"
    )
    parser.add_argument("--mine", action="store_true", default=False)
    parser.add_argument("key")


def create_table() -> Table:
    """Create table for displaying ContentFiles"""
    table = Table(box=box.ROUNDED, style="box")
    table.add_column("Size", justify="right", header_style="header")
    table.add_column("Timestamp", header_style="header")
    table.add_column("Package", header_style="header", overflow="fold")
    table.add_column("Path", header_style="header", overflow="fold")

    return table


def format_content_file(
    content_file: dict[str, Any], args: argparse.Namespace
) -> tuple[str, ...]:
    """Pretty-format the given content file giving args"""
    cf = SimpleNamespace(**content_file)
    binpkg = SimpleNamespace(**cf.binpkg)
    build = SimpleNamespace(**binpkg.build)
    machine = render.format_machine(build.machine, args)
    build_id = build.id.rsplit(".")[-1]
    timestamp = render.format_timestamp(
        dt.datetime.fromisoformat(cf.timestamp).astimezone(render.LOCAL_TIMEZONE)
    )
    return (
        f"[filesize]{cf.size}[/filesize]",
        f"[timestamp]{timestamp}[/timestamp]",
        (
            f"[machine]{machine}[/machine]"
            f"/[build_id]{build_id}[/build_id]"
            f"/[package]{binpkg.cpvb}"
        ),
        f"[tag]{cf.path}[/tag]",
    )



"""The gbp fl ls subcommand"""

import argparse
import datetime as dt
from typing import TypedDict

from gbpcli import render
from gbpcli.gbp import GBP
from gbpcli.types import Console
from rich import box
from rich.table import Table

from gbp_fl import utils

from . import resolve_build_id

HELP = "List files in a package"


class ContentFileDict(TypedDict):
    """ContentFiles dict returned from queries"""

    path: str
    size: int
    timestamp: str


def handler(args: argparse.Namespace, gbp: GBP, console: Console) -> int:
    """List files in a package"""
    if (spec := utils.parse_pkgspec(args.pkgspec)) is None:
        console.err.print(f"[red]Invalid specifier: {args.pkgspec}[/red]")
        return 1

    build_id = resolve_build_id(spec.machine, spec.build_id, gbp) or ""
    response, _ = gbp.query.gbp_fl.list(  # type: ignore
        machine=spec.machine, buildId=build_id, cpvb=spec.cpvb, extended=args.long
    )
    fl_list: list[ContentFileDict] = response["flList"]
    fl_list.sort(key=lambda item: item["path"])

    (print_long if args.long else print_short)(fl_list, console)

    return 0


def parse_args(parser: argparse.ArgumentParser) -> None:
    """Build command-line arguments"""
    parser.add_argument(
        "-l", "--long", action="store_true", default=False, help="Print in long format"
    )
    parser.add_argument("pkgspec")


def print_short(cfs: list[ContentFileDict], console: Console) -> None:
    """Print ContentFiles in short format"""
    for item in cfs:
        console.out.print(item["path"])


def print_long(cfs: list[ContentFileDict], console: Console) -> None:
    """Print ContentFiles in long format"""
    table = create_table()
    for item in cfs:
        table.add_row(*format_row(item))

    console.out.print(table)


def create_table() -> Table:
    """Create and return a Table for displaying files in long format"""
    table = Table(box=box.ROUNDED, style="box")
    table.add_column("Size", justify="right", header_style="header")
    table.add_column("Timestamp", header_style="header")
    table.add_column("Path", header_style="header")

    return table


def format_row(item: ContentFileDict) -> tuple[str, str, str]:
    """Format the item for adding to a table"""
    timestamp = render.format_timestamp(
        dt.datetime.fromisoformat(item["timestamp"]).astimezone(render.LOCAL_TIMEZONE)
    )
    return (
        f"[filesize]{item['size']}[/filesize]",
        f"[timestamp]{timestamp}[/timestamp]",
        f"[tag]{item['path']}[/tag]",
    )



"""gbp-fl's command-line interface

The gbpcli handler does nothing but print usage as `gbp fl` has subparsers itself and
there is nothing (yet) to do at the "root" level.
"""

import argparse
import functools
import importlib
from typing import IO

from gbpcli.gbp import GBP
from gbpcli.types import Console

HELP = "Access the Gentoo Build Publisher File List"
SUBCOMMANDS = ("fetch", "ls", "search")


def handler(args: argparse.Namespace, _gbp: GBP, console: Console) -> int:
    """GBP File List"""
    print_usage(console.err.file)
    return 1


def parse_args(parser: argparse.ArgumentParser) -> None:
    """Create parsers for subcommands"""
    subparsers = parser.add_subparsers()

    for subcommand in SUBCOMMANDS:

        module = importlib.import_module(f"gbp_fl.cli.{subcommand}")
        subparser = subparsers.add_parser(subcommand, help=module.HELP)
        module.parse_args(subparser)
        subparser.set_defaults(func=getattr(module, "handler"))


def print_usage(io: IO[str]) -> None:
    """Print the usage to the given file handle"""
    uprint = functools.partial(print, file=io)
    uprint("Subcommands:")

    for subcommand in SUBCOMMANDS:
        uprint(f"  {subcommand}")


def resolve_build_id(machine: str, build_id_or_tag: str, gbp: GBP) -> str | None:
    """Return build_id or the the build_id for the given tag

    If the given tag does not resolve to a build, return None.
    """
    if not build_id_or_tag.startswith("@"):
        return build_id_or_tag

    build = gbp.resolve_tag(machine, build_id_or_tag[1:])

    return str(build.number) if build else None



# pylint: disable=missing-docstring,unused-argument
import datetime as dt
from argparse import ArgumentParser
from dataclasses import replace
from unittest import TestCase
from unittest.mock import Mock, patch

import gbp_testkit.fixtures as testkit
from gbp_testkit.helpers import parse_args, print_command
from unittest_fixtures import Fixtures, given, where

from gbp_fl.cli import search
from gbp_fl.types import ContentFile

from . import lib

DAY = dt.timedelta(days=1, minutes=11, seconds=12)


@given(lib.environ, lib.gbp_client, lib.repo, lib.bulk_content_files, testkit.console)
@given(lib.local_timezone)
@where(repo="gbp_fl.graphql.queries.repo", environ={"GBPCLI_MYMACHINES": "lighthouse"})
@patch("gbp_fl.graphql.binpkg.gateway")
class SearchTests(TestCase):
    options = {"records_backend": "memory"}

    def test(self, gateway: Mock, fixtures: Fixtures) -> None:
        cfs = fixtures.bulk_content_files
        repo = fixtures.repo
        now = fixtures.now

        for cf in cfs:
            cf = replace(cf, timestamp=now)
            repo.files.save(cf)
            now = now + DAY

        bash_file_indexes = [0, 3, 4, 5]

        make_build_records(gateway, [cfs[i] for i in bash_file_indexes])

        cmdline = "gbp fl search bash"
        args = parse_args(cmdline)
        gbp = fixtures.gbp_client
        console = fixtures.console

        print_command(cmdline, console)
        status = search.handler(args, gbp, console)

        self.assertEqual(status, 0)
        self.assertEqual(
            TEST1_SEARCH_OUTPUT,
            console.out.file.getvalue(),
            "\n" + console.out.file.getvalue(),
        )

    def test_with_machine(self, gateway: Mock, fixtures: Fixtures) -> None:
        cfs = fixtures.bulk_content_files
        repo = fixtures.repo
        now = fixtures.now

        for cf in cfs:
            cf = replace(cf, timestamp=now)
            repo.files.save(cf)
            now = now + DAY

        bash_file_indexes = [0, 3, 4, 5]

        make_build_records(gateway, [cfs[i] for i in bash_file_indexes])

        cmdline = "gbp fl search -m lighthouse bash"
        args = parse_args(cmdline)
        gbp = fixtures.gbp_client
        console = fixtures.console

        print_command(cmdline, console)
        status = search.handler(args, gbp, console)

        self.assertEqual(status, 0)
        self.assertEqual(
            TEST3_SEARCH_OUTPUT,
            console.out.file.getvalue(),
            "\n" + console.out.file.getvalue(),
        )

    def test_with_mine(self, gateway: Mock, fixtures: Fixtures) -> None:
        cfs = fixtures.bulk_content_files
        repo = fixtures.repo
        now = fixtures.now

        for cf in cfs:
            cf = replace(cf, timestamp=now)
            repo.files.save(cf)
            now = now + DAY

        bash_file_indexes = [0, 3, 4, 5]

        make_build_records(gateway, [cfs[i] for i in bash_file_indexes])

        cmdline = "gbp fl search --mine bash"
        args = parse_args(cmdline)
        gbp = fixtures.gbp_client
        console = fixtures.console

        print_command(cmdline, console)
        status = search.handler(args, gbp, console)

        self.assertEqual(status, 0)
        self.assertEqual(
            TEST4_SEARCH_OUTPUT,
            console.out.file.getvalue(),
            "\n" + console.out.file.getvalue(),
        )

    def test_no_match(self, gateway: Mock, fixtures: Fixtures) -> None:
        cmdline = "gbp fl search bash"
        args = parse_args(cmdline)
        gbp = fixtures.gbp_client
        console = fixtures.console

        status = search.handler(args, gbp, console)

        self.assertEqual(status, 0)
        self.assertEqual("", console.out.file.getvalue())

    def test_with_missing_metadata(self, gateway: Mock, fixtures: Fixtures) -> None:
        # When a package is being deleted, e.g. during a purge run, the binpkg metadata
        # may be gone. This results in an error in the graphql call and the binpkg field
        # being set to null. We want to (silently) ignore those
        cfs = fixtures.bulk_content_files
        repo = fixtures.repo
        repo.files.bulk_save(cfs)

        bash_file_indexes = [0, 3, 4, 5]
        build_records = [
            Mock(
                machine=cfs[i].binpkg.build.machine,
                build_id=cfs[i].binpkg.build.build_id,
                id=cfs[i].binpkg.build.build_id,
            )
            for i in bash_file_indexes
        ]
        # make the second one return None
        build_records[1] = None  # type: ignore
        gateway.get_build_record.side_effect = tuple(build_records)

        cmdline = "gbp fl search bash"
        args = parse_args(cmdline)
        gbp = fixtures.gbp_client
        console = fixtures.console

        print_command(cmdline, console)
        status = search.handler(args, gbp, console)

        self.assertEqual(status, 0)
        self.assertEqual(
            TEST2_SEARCH_OUTPUT,
            console.out.file.getvalue(),
            "\n" + console.out.file.getvalue(),
        )


class ParseArgsTests(TestCase):
    def test(self) -> None:
        parser = ArgumentParser()
        search.parse_args(parser)


def make_build_records(gateway: Mock, content_files: list[ContentFile]) -> None:
    """Mock gateway to return content_file when .get_build_record() is called"""
    gateway.get_build_record.side_effect = tuple(
        Mock(
            machine=content_file.binpkg.build.machine,
            build_id=content_file.binpkg.build.build_id,
            id=content_file.binpkg.build.build_id,
        )
        for content_file in content_files
    )


TEST1_SEARCH_OUTPUT = """$ gbp fl search bash
╭────────┬───────────────────┬─────────────────────────────────────────┬───────────╮
│   Size │ Timestamp         │ Package                                 │ Path      │
├────────┼───────────────────┼─────────────────────────────────────────┼───────────┤
│ 850648 │ 01/26/25 05:57:37 │ lighthouse/34/app-shells/bash-5.2_p37-1 │ /bin/bash │
│ 850648 │ 01/29/25 06:31:13 │ polaris/26/app-shells/bash-5.2_p37-1    │ /bin/bash │
│ 850648 │ 01/30/25 06:42:25 │ polaris/26/app-shells/bash-5.2_p37-2    │ /bin/bash │
│ 850648 │ 01/31/25 06:53:37 │ polaris/27/app-shells/bash-5.2_p37-1    │ /bin/bash │
╰────────┴───────────────────┴─────────────────────────────────────────┴───────────╯
"""
TEST2_SEARCH_OUTPUT = """$ gbp fl search bash
╭────────┬───────────────────┬─────────────────────────────────────────┬───────────╮
│   Size │ Timestamp         │ Package                                 │ Path      │
├────────┼───────────────────┼─────────────────────────────────────────┼───────────┤
│ 850648 │ 01/26/25 05:57:37 │ lighthouse/34/app-shells/bash-5.2_p37-1 │ /bin/bash │
│ 850648 │ 01/26/25 05:57:37 │ polaris/26/app-shells/bash-5.2_p37-2    │ /bin/bash │
│ 850648 │ 01/26/25 05:57:37 │ polaris/27/app-shells/bash-5.2_p37-1    │ /bin/bash │
╰────────┴───────────────────┴─────────────────────────────────────────┴───────────╯
"""
TEST3_SEARCH_OUTPUT = """$ gbp fl search -m lighthouse bash
╭────────┬───────────────────┬─────────────────────────────────────────┬───────────╮
│   Size │ Timestamp         │ Package                                 │ Path      │
├────────┼───────────────────┼─────────────────────────────────────────┼───────────┤
│ 850648 │ 01/26/25 05:57:37 │ lighthouse/34/app-shells/bash-5.2_p37-1 │ /bin/bash │
╰────────┴───────────────────┴─────────────────────────────────────────┴───────────╯
"""
TEST4_SEARCH_OUTPUT = """$ gbp fl search --mine bash
╭────────┬───────────────────┬─────────────────────────────────────────┬───────────╮
│   Size │ Timestamp         │ Package                                 │ Path      │
├────────┼───────────────────┼─────────────────────────────────────────┼───────────┤
│ 850648 │ 01/26/25 05:57:37 │ lighthouse/34/app-shells/bash-5.2_p37-1 │ /bin/bash │
╰────────┴───────────────────┴─────────────────────────────────────────┴───────────╯
"""
